"""Functions to interact with a webdriver browser object."""
# import time
# import types
import traceback
from contextlib import contextmanager

from selenium import webdriver
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities

from golem.core import utils
from golem import execution
from golem.webdriver import (GolemChromeDriver,
                             GolemEdgeDriver,
                             GolemGeckoDriver,
                             GolemIeDriver,
                             GolemOperaDriver,
                             GolemRemoteDriver)
from golem.webdriver.extended_driver import GolemExtendedDriver


def element(*args, **kwargs):
    """Shortcut to golem.browser.get_browser().find()"""
    webelement = get_browser().find(*args, **kwargs)
    return webelement


def elements(*args, **kwargs):
    """Shortcut to golem.browser.get_browser().find_all()"""
    webelement = get_browser().find_all(*args, **kwargs)
    return webelement


def open_browser(browser_id=None):
    """Open a browser.

    When opening more than one browser instance per test
    provide a browser_id to switch between browsers later on
    """
    @contextmanager
    def validate_exec_path(browser_name, exec_path_setting, settings):
        executable_path = settings[exec_path_setting]
        if executable_path:
            matched_executable_path = utils.match_latest_executable_path(executable_path)
            if matched_executable_path:
                try:
                    yield matched_executable_path
                except:
                    msg = ('Could not start {} driver using the path \'{}\'\n'
                           'verify that the {} setting points to a valid webdriver executable.'
                           .format(browser_name, executable_path, exec_path_setting))
                    execution.logger.error(msg)
                    execution.logger.info(traceback.format_exc())
                    raise Exception(msg)
            else:
                msg = 'No executable file found using path {}'.format(executable_path)
                execution.logger.error(msg)
                raise Exception(msg)
        else:
            msg = '{} setting is not defined'.format(exec_path_setting)
            execution.logger.error(msg)
            raise Exception(msg)
    
    @contextmanager
    def validate_remote_url(remote_url):
        if remote_url:
            yield remote_url
        else:
            msg = 'remote_url setting is required'
            execution.logger.error(msg)
            raise Exception(msg)

    driver = None
    browser_definition = execution.browser_definition
    settings = execution.settings
    # remote
    if browser_definition['remote'] is True:
        with validate_remote_url(settings['remote_url']) as remote_url:
            driver = GolemRemoteDriver(command_executor=remote_url,
                                       desired_capabilities=browser_definition['capabilities'])
    # Chrome
    elif browser_definition['name'] == 'chrome':
        with validate_exec_path('chrome', 'chromedriver_path', settings) as ex_path:
            chrome_options = webdriver.ChromeOptions()
            chrome_options.add_argument('--start-maximized')
            driver = GolemChromeDriver(executable_path=ex_path,
                                       chrome_options=chrome_options)
    # Chrome remote
    elif browser_definition['name'] == 'chrome-remote':
        with validate_remote_url(settings['remote_url']) as remote_url:
            driver = GolemRemoteDriver(command_executor=remote_url,
                                       desired_capabilities=DesiredCapabilities.CHROME)
    # Chrome headless
    elif browser_definition['name'] == 'chrome-headless':
        with validate_exec_path('chrome', 'chromedriver_path', settings) as ex_path:
            chrome_options = webdriver.ChromeOptions()
            chrome_options.add_argument('headless')
            chrome_options.add_argument('--window-size=1600,1600')
            driver = GolemChromeDriver(executable_path=ex_path,
                                       chrome_options=chrome_options)
    # Chrome remote headless
    elif browser_definition['name'] == 'chrome-remote-headless':
        with validate_remote_url(settings) as remote_url:
            chrome_options = webdriver.ChromeOptions()
            chrome_options.add_argument('headless')
            desired_capabilities = chrome_options.to_capabilities()
            driver = GolemChromeDriver(command_executor=remote_url,
                                       desired_capabilities=desired_capabilities)
    # Edge
    elif browser_definition['name'] == 'edge':
        with validate_exec_path('edge', 'edgedriver_path', settings) as ex_path:
            driver = GolemEdgeDriver(executable_path=ex_path)        
    # Edge remote
    elif browser_definition['name'] == 'edge-remote':
        with validate_remote_url(settings['remote_url']) as remote_url:
            driver = GolemRemoteDriver(command_executor=remote_url,
                                       desired_capabilities=DesiredCapabilities.EDGE)
    # Firefox
    elif browser_definition['name'] == 'firefox':
        with validate_exec_path('firefox', 'geckodriver_path', settings) as ex_path:
            driver = GolemGeckoDriver(executable_path=ex_path)
    # Firefox remote
    elif browser_definition['name'] == 'firefox-remote':
        with validate_remote_url(settings['remote_url']) as remote_url:
            driver = GolemRemoteDriver(command_executor=remote_url,
                                       desired_capabilities=DesiredCapabilities.FIREFOX)
    # IE
    elif browser_definition['name'] == 'ie':
        with validate_exec_path('internet explorer', 'iedriver_path', settings) as ex_path:
            driver = GolemIeDriver(executable_path=ex_path)
    # IE remote
    elif browser_definition['name'] == 'ie-remote':
        with validate_remote_url(settings['remote_url']) as remote_url:
            driver = GolemRemoteDriver(command_executor=remote_url,
                                       desired_capabilities=DesiredCapabilities.IE)
    # Opera
    elif browser_definition['name'] == 'opera':
        with validate_exec_path('opera', 'operadriver_path', settings) as ex_path:
            opera_options = webdriver.ChromeOptions()
            if 'opera_binary_path' in settings:
                opera_options.binary_location = settings['opera_binary_path']
            driver = GolemOperaDriver(executable_path=ex_path, options=opera_options)
    # Opera remote
    elif browser_definition['name'] == 'opera-remote':
        with validate_remote_url(settings['remote_url']) as remote_url:
            driver = GolemRemoteDriver(command_executor=remote_url,
                                       desired_capabilities=DesiredCapabilities.OPERA)
    # Safari
    # elif browser_definition['name'] == 'safari':
    #     with validate_exec_path('safari', 'safaridriver_path', settings) as ex_path:
    #         driver = GolemSafariDriver(executable_path=ex_path)
    # Safari remote
    # elif browser_definition['name'] == 'safari-remote':
    #     with validate_remote_url(settings['remote_url']) as remote_url:
    #         driver = GolemRemoteDriver(command_executor=remote_url,
    #                                   desired_capabilities=DesiredCapabilities.SAFARI)
    else:
        raise Exception('Error: {} is not a valid driver'.format(browser_definition['name']))

    driver.maximize_window()

    if not browser_id:
        if len(execution.browsers) == 0:
            browser_id = 'main'
        else:
            browser_id = 'browser{}'.format(len(execution.browsers))
    execution.browsers[browser_id] = driver
    if not execution.browser:
        execution.browser = driver


def get_browser() -> GolemRemoteDriver:
    """Returns the active browser. Starts a new one if there is none."""
    if not execution.browser:
        open_browser()
    return execution.browser


def activate_browser(browser_id):
    """Activate a browser.
    Only needed when the test starts more than one browser instance."""
    if not browser_id in execution.browsers:
        # TODO, use error() function
        raise Exception("Error: {} is not a valid browser id. Current browsers "
                        "are: {}".format(browser_id, ', '.join(execution.browsers.keys())))
    else:
        execution.browser = execution.browsers[browser_id]
